#pragma once

#include <Base_Receiver.h>
#include <Base_Configurator.h>

/**
 * @brief The namespace inaf::oasbo::Receivers contains classes related to receivers in the ASTRI DAQ system.
 */
namespace inaf::oasbo::Receivers {

/**
 * @brief The TCPProtocol class is a derived class of BaseReceiver and represents a TCP protocol receiver.
 *  check the Base_Receiver.h file for more information.
 */
class TCPProtocol: public BaseReceiver {
protected:
	/**
	 * @brief Constructs a TCPProtocol object with the specified IP address and port number.
	 * @param ip The IP address to connect to.
	 * @param prt The port number to connect to.
	 */
	TCPProtocol(std::string ip, int prt);

	/**
	 * @brief Constructs a TCPProtocol object with default values for IP address and port number.
	 */
	TCPProtocol();

	std::string ip; /**< The IP address to connect to. */
	int port; /**< The port number to connect to. */
	int srv_sock; /**< The server socket. */
	int cli_sock; /**< The client socket. */
	struct sockaddr_in servaddr; /**< The server address. */
	struct sockaddr_in cliaddr; /**< The client address. */

	bool checkConnFlag = true; /**< Flag that tells if it has to check for incoming connection. Set to false when calling closeConnectionToClient */

	/**
	 * @brief Helper function. Connects to the client with the specified IP address and port number.
	 * @param ip The IP address to connect to.
	 * @param port The port number to connect to.
	 * @return The result of the connection attempt.
	 */
	int m_connectToCli(std::string ip, int port);

	/**
	 * @brief Resets the first bytes of the packet to 0.
	 * @param pack The packet to reset.
	 * @param bytes The number of bytes to reset.
	 */
	void resetPacket(inaf::oasbo::Packets::BasePacket &pack, int bytes);

	/**
	 * @brief Splits the IP address and port number from the specified string.
	 * @param ip_port The string containing the IP address and port number.
	 * @param ip The extracted IP address.
	 * @param port The extracted port number.
	 * @return True if the split was successful, false otherwise.
	 */
	bool split_ip_port(const std::string &ip_port, std::string &ip, int &port);

public:
	/**
	 * @brief Destroys the TCPProtocol object.
	 */
	~TCPProtocol();

	/**
	 * @brief Gets the host name.
	 * @return The host name.
	 */
	std::string getHost() override;

	/**
	 * @brief Sets the host name.
	 * @param host The host name to set.
	 */
	void setHost(std::string host) override;

	/**
	 * @brief Sets the IP address.
	 * @param ip The IP address to set.
	 */
	void setIp(std::string);

	/**
	 * @brief Gets the IP address.
	 * @return The IP address.
	 */
	std::string getIp() {
		return this->ip;
	}

	/**
	 * @brief Sets the port number.
	 * @param port The port number to set.
	 */
	void setPort(int);

	/**
	 * @brief Gets the port number.
	 * @return The port number.
	 */
	int getPort() {
		return this->port;
	}

	int connectToClient() override;

	int closeConnectionToClient() override;

	bool isConnectedToClient() const override;

	int receiveFromClient(Packets::BasePacket&) override;

	friend class TCPProtocolBuilder;
};

/**
 * @brief The TCPProtocolBuilder class is a builder class for TCPProtocol objects.
 */
class TCPProtocolBuilder {
protected:
	TCPProtocol *protocol; /**< The TCPProtocol object being built. */

public:
	std::string config_target { "tcpreceiver" }; /**< The configuration target. */
	std::string ip_key { "ip" }; /**< The IP configuration key. */
	std::string port_key { "port" }; /**< The port configuration key. */

	/**
	 * @brief Constructs a TCPProtocolBuilder object.
	 */
	TCPProtocolBuilder();

	/**
	 * @brief Constructs a TCPProtocolBuilder object with the specified IP address and port number.
	 * @param ip The IP address to set.
	 * @param port The port number to set.
	 */
	TCPProtocolBuilder(std::string ip, int port);

	/**
	 * @brief Destroys the TCPProtocolBuilder object.
	 */
	~TCPProtocolBuilder();

	/**
	 * @brief Resets the builder to its initial state.
	 */
	void reset();

	/**
	 * @brief Configures the builder from the specified configurator.
	 * @param conf The configurator to use.
	 * @return The configured TCPProtocolBuilder object.
	 */
	TCPProtocolBuilder* configFrom(Configurators::BaseConfigurator &conf);

	/**
	 * @brief Sets the IP address.
	 * @param ip The IP address to set.
	 * @return The TCPProtocolBuilder object.
	 */
	TCPProtocolBuilder* setIp(std::string ip);

	/**
	 * @brief Sets the port number.
	 * @param port The port number to set.
	 * @return The TCPProtocolBuilder object.
	 */
	TCPProtocolBuilder* setPort(int port);

	/**
	 * @brief Gets the built TCPProtocol object.
	 * @return The built TCPProtocol object.
	 */
	TCPProtocol* getReceiver();
};

} // namespace inaf::oasbo::Receivers
