#include <iostream>
#include <string>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <netdb.h>
#include <regex>
#include <TCP_Protocol.h>

#include <ctime>
#include <iomanip>
#include <fcntl.h>
#include <chrono>
#include <thread>

using namespace inaf::oasbo::Receivers;

TCPProtocol::TCPProtocol() :
		TCPProtocol("localhost", 9003) {
}
TCPProtocol::TCPProtocol(std::string ip, int prt) :
		ip(ip), port(prt), srv_sock(-1), cli_sock(-1) {
	host = ip + ":" + std::to_string(port);
	memset(&cliaddr, 0, sizeof(cliaddr));
	memset(&servaddr, 0, sizeof(servaddr));
}

TCPProtocol::~TCPProtocol() {
	closeConnectionToClient();
	if (srv_sock != -1) {
		::close(srv_sock);
	}
}

int TCPProtocol::receiveFromClient(Packets::BasePacket &pack) {
	ssize_t headerSize = pack.getHeaderSize();
	uint8_t *buff = new uint8_t[pack.getPacketStructureByteSize()];
	ssize_t rec = 0;

	ssize_t to_be_rcv = headerSize;
	while (this->checkConnFlag && to_be_rcv > 0) { // go ahead until headersize received or connection terminates
		rec = ::recv(this->cli_sock, &buff[0], to_be_rcv,
		MSG_DONTWAIT);
		if (rec == 0) { // connection closed
			closeConnectionToClient();
			delete[] buff;
			return 0;
		}
		if (rec == -1) {
			if (errno == EAGAIN || errno == EWOULDBLOCK) {
				continue; // No data available
			} else {
				return -1;
			}
		}
		to_be_rcv -= rec;
	}

	if (!this->checkConnFlag) { // the connection has been terminated
		closeConnectionToClient();
		delete[] buff;
		return 0;
	}

	// entire header received...
	pack.copyToMemory(buff, headerSize);
	if (!pack.hasRecognizedHeader()) { // header not recognized
		resetPacket(pack, headerSize);
		delete[] buff;
		return -1;
	}

	// receive the rest of the packet
	size_t payload_and_tail_size = pack.getPayloadSize() + pack.getTailSize();
	to_be_rcv = payload_and_tail_size;
	while (this->checkConnFlag && to_be_rcv > 0) { // go ahead until the rest of the packet received or connection terminates
		rec = ::recv(this->cli_sock, &buff[headerSize], to_be_rcv,
		MSG_DONTWAIT);
		if (rec == 0) { // connection closed
			closeConnectionToClient();
			delete[] buff;
			return 0;
		}
		if (rec == -1) {
			if (errno == EAGAIN || errno == EWOULDBLOCK) {
				continue; // No data available
			} else {
				return -1;
			}
		}
		to_be_rcv -= rec;
	}

	if (!this->checkConnFlag) { // the connection has been terminated
		closeConnectionToClient();
		delete[] buff;
		return 0;
	}

	pack.copyToMemory(&buff[headerSize], payload_and_tail_size, headerSize);
	delete[] buff;

	return headerSize + payload_and_tail_size;
}

int TCPProtocol::connectToClient() {
	this->checkConnFlag = true;
	return m_connectToCli(ip, port);
}

int TCPProtocol::closeConnectionToClient() {
	this->checkConnFlag = false;
	if (cli_sock != -1) {
		::close(cli_sock);
		cli_sock = -1;
	}
	return 1;
}

bool TCPProtocol::isConnectedToClient() const {
	return cli_sock > 0;
}

int TCPProtocol::m_connectToCli(std::string ip, int port) {
	// Creating socket file descriptor
	if (srv_sock == -1) { // bind
		if ((srv_sock = socket(AF_INET, SOCK_STREAM, 0)) == 0) {
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[TCP Receiver]\t" << "TCP socket on " << host
					<< " failed" << std::endl;
			return -1;
		}
		servaddr.sin_family = AF_INET;
		servaddr.sin_addr.s_addr = ::inet_addr(ip.c_str());
		servaddr.sin_port = htons(port);
		if (bind(srv_sock, (struct sockaddr*) &servaddr, sizeof(servaddr))
				< 0) {
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[TCP Receiver]\t" << "TCP CONNECTION on " << host
					<< " Bind failed" << std::endl;
			srv_sock = -1;
			return -1;
		}
	}

	if (listen(srv_sock, 1) < 0) {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[TCP Receiver]\t" << "Listen error" << std::endl;
		srv_sock = -1;
		return -1;
	}

	time_t now = time(nullptr);
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[TCP Receiver]\t" << "Waiting for connection on " << host
			<< std::endl;

	while (this->checkConnFlag) {
		fd_set read_fds;
		FD_ZERO(&read_fds);
		FD_SET(srv_sock, &read_fds);

		struct timeval timeout;
		timeout.tv_sec = 1;
		timeout.tv_usec = 0;
		int select_status = select(srv_sock + 1, &read_fds, NULL, NULL,
				&timeout);
		if (select_status > 0) {
			break;  // we have data, we can accept now
		}
	}

	if (!this->checkConnFlag) { // Interrupted
		closeConnectionToClient();
		return 0;
	}

	socklen_t clientAddressSize = sizeof(cliaddr);
	if ((cli_sock = accept(srv_sock, (struct sockaddr*) &cliaddr,
			(socklen_t*) &clientAddressSize)) < 0) {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[TCP Receiver]\t" << "accept error" << std::endl;
		cli_sock = -1;
		return -1;
	}
	now = time(nullptr);
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[TCP Receiver]\t" << "Connected with "
			<< inet_ntoa(cliaddr.sin_addr) << std::endl;
	return cli_sock;
}

void TCPProtocol::resetPacket(Packets::BasePacket &pack, int bytes) {
	uint8_t *buff = new uint8_t[bytes];
	std::memset(buff, 0, bytes);
	int toBeReset = std::min(
			static_cast<int>(pack.getPacketStructureByteSize()), bytes);
	pack.copyToMemory(buff, toBeReset);
	delete[] buff;
}

std::string TCPProtocol::getHost() {
	return ip + ":" + std::to_string(port);
}

void TCPProtocol::setHost(std::string host) {
	std::string ip = { };
	int port = 0;
	if (split_ip_port(host, ip, port)) {
		setIp(ip);
		setPort(port);
	} else {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[TCP Receiver]\t"
				<< "invalid IP address and port format: " << host << std::endl;
	}
}

void TCPProtocol::setIp(std::string ip) {
	std::regex pattern("^[\\w.-]+$");
	// Check if the input string matches the expected format
	bool ok = true;
	if (!std::regex_match(ip, pattern)) {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[TCP Receiver]\t" << "invalid IP address: " << ip
				<< ", setting to localhost" << std::endl;
		ok = false;
		this->ip = std::string("localhost");
	}
	if (ok)
		this->ip = ip;

	this->host = this->ip + ":" + std::to_string(port);
}
void TCPProtocol::setPort(int port) {
	if (port > 1023 && port < 65535)
		this->port = port;
	else {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[TCP Receiver]\t" << "invalid port: " << port
				<< ", setting to 9003" << std::endl;
		this->port = 9003;
	}
	this->host = this->ip + ":" + std::to_string(this->port);
}

bool TCPProtocol::split_ip_port(const std::string &ip_port,
		std::string &ip_address, int &port) {
	// Regex pattern to match IP address and port in the format "xxx.xxx.xxx.xxx:xxxx"
	std::regex pattern("^[\\w.-]+:(\\d{1,5})$");

	// Check if the input string matches the expected format
	if (!std::regex_match(ip_port, pattern)) {
		return false;
	}

	// Split the input string into IP address and port
	int colon_pos = ip_port.find(":");
	ip_address = ip_port.substr(0, colon_pos);
	port = std::stoi(ip_port.substr(colon_pos + 1));
	return true;
}
